#include "MolReader.h"

using namespace std;
using namespace compound;
using namespace utils;

bool MolReader::readMol(Molecule *mol, string str)
{
    istringstream in(str);
    return read(mol, in);
}


bool MolReader::readMolFile(Molecule *mol, string file)
{
    ifstream in;
    in.open(file.c_str(), ios::in);

    if(!in.good())
    {
        cerr << "Cound not open " << file << endl;
        return false;
    }

    return read(mol, in);
}

bool MolReader::read(Molecule *mol, istream &in)
{
    string line;

    // read header (3lines)
    getline(in, line);
    getline(in, line);
    getline(in, line);

    // read atom and bond number.
    if(!getline(in, line))
    {
        cerr << "Illegal counts line: Counts line not found." << endl;
        return false;
    }

    if(line.length() < 6)
    {
        cerr << "Illegal counts line: Number of atoms, bonds not found." << endl;
        cerr << line << endl;
        return false;
    }

    int atomNumber = atoi(line.substr(0, 3).c_str());
    int bondNumber = atoi(line.substr(3, 3).c_str());

    // read atom block
    if(!readAtoms(mol, in, atomNumber))
        return false;

    // read bond block
    if(!readBonds(mol, in, bondNumber))
        return false;

    if(!readProperties(mol, in))
        return false;

    return true;
}


bool MolReader::readAtoms(Molecule *mol, istream &in, int atomNumber)
{
    string line;

    for(int i=0; i<atomNumber; i++)
    {
        if(!getline(in, line))
            return false;

        double x, y, z;
        int mass, charge, scan;
        char s[4];
        memset(s, 0x00, sizeof(s));

        scan = sscanf(line.c_str(),
                "%10lf%10lf%10lf %3s%2d%3d",
                &x, &y, &z, s, &mass, &charge);

        if(scan != 6)
        {
            cerr << "Illegal atom block:" << endl << line << endl;
            return false;
        }


        if(charge == 1)
            charge = 3;
        else if(charge == 2)
            charge = 2;
        else if(charge == 3)
            charge = 1;
        else if(charge == 5)
            charge = -1;
        else if(charge == 6)
            charge = -2;
        else if(charge == 7)
            charge = -3;
        else
            charge = 0;


        Atom *atom = mol->addNewAtom();

        atom->setNumber((i+1));
        atom->setSymbol(string(s));
        atom->setX(x);
        atom->setY(y);
        atom->setZ(z);
        atom->setCharge(charge);
    }

    return true;
}


bool MolReader::readBonds(Molecule *mol, istream &in, int bondNumber)
{
    string line;

    for(int i=0; i<bondNumber; i++)
    {
        if(!getline(in, line))
            return false;


        if(line.length() < 12)
        {
            cerr << "Illegal bond block:" << endl << line << endl;
            return false;
        }

        int firstAtomNum, secondAtomNum, type, stereo;

        firstAtomNum = atoi(line.substr(0, 3).c_str());
        secondAtomNum = atoi(line.substr(3, 3).c_str());
        type = atoi(line.substr(6, 3).c_str());
        stereo = atoi(line.substr(9, 3).c_str());

        Atom *firstAtom, *secondAtom;
        firstAtom = mol->getAtom(firstAtomNum - 1);
        secondAtom = mol->getAtom(secondAtomNum - 1);

        if(!firstAtom || !secondAtom)
        {
            cout << firstAtomNum << ", " << secondAtomNum << endl;
            cerr << "Illegal bond block: first atom or second atom not found." << endl;
            cerr << line << endl;
            return false;
        }

        Bond *bond = mol->addNewBond();

        bond->setFirstAtom(firstAtom);
        bond->setSecondAtom(secondAtom);
        bond->setType(type);
        bond->setStereo(stereo);

        firstAtom->addBond(bond);
        secondAtom->addBond(bond);
    }

    return true;
}


bool MolReader::readProperties(Molecule *mol, istream &in)
{
    string line;

    while(getline(in, line))
    {
        if(line.find("M  CHG") == 0)
            readCharge(mol, line);
        else if(line.find("M  ISO") == 0)
            readIsotope(mol, line);
        else if(line.find("M  STY") == 0)
            readSgroupType(mol, line);
        else if(line.find("M  SMT") == 0)
            readSgroupSubscript(mol, line);
        else if(line.find("M  SDI") == 0)
            readSgroupDisplayInformation(mol, line);
    }

    return true;
}


bool MolReader::readCharge(Molecule *mol, string line)
{
    // line = "M  CHG  5   1   4   2   -1...."

    istringstream iss(line);

    string buff;
    int num;

    iss >> buff;  // "M"
    iss >> buff;  // "CHG"
    iss >> num;   // "5"

    int atomNum, charge;
    for(int i=0; i<num; i++)
    {
        iss >> atomNum >> charge;  // "1   4"

        Atom *atom = mol->getAtom((atomNum - 1));
        if(!atom)
        {
            cerr << "Illegal 'M  CHG' block: Atom not found." << endl;
            cerr << line << endl;
            return false;
        }
        else
        {
            atom->setCharge(charge);
        }
    }

    return true;
}


bool MolReader::readIsotope(Molecule *mol, string line)
{
    // line = "M  ISO  1  36 125"

    istringstream iss(line);
    string buff;
    int num;

    iss >> buff;  // "M"
    iss >> buff;  // "ISO"
    iss >> num;

    for(int i=0; i<num; i++)
    {
        int atomNum, isotope;
        iss >> atomNum;
        iss >> isotope;

        Atom *atom = mol->getAtom((atomNum - 1));
        if(!atom)
        {
            cerr << "Illegal 'M  ISO' block: Atom not found." << endl;
            cerr << line << endl;
            return false;
        }
        else
        {
            atom->setIsotope(isotope);
        }
    }

    return true;
}


bool MolReader::readSgroupType(Molecule *mol, string line)
{
    int num = 0;
    sscanf(line.c_str(), "M  STY %d", &num);

    while(num > 0)
    {
        mol->addNewBracket();
        --num;
    }

    return true;
}


bool MolReader::readSgroupSubscript(Molecule *mol, string line)
{
    int num, scan;
    char subscript[256];
    memset(subscript, 0x00, sizeof(subscript));

    scan = sscanf(line.c_str(), "M  SMT %d %s", &num, subscript);

    if(scan != 2)
    {
        cerr << "Illegal SMT line:" << endl;
        cerr << line << endl;
        return false;
    }

    Bracket *bracket = mol->getBracket(num - 1);
    if(bracket)
    {
        bracket->setSubscript(string(subscript));
    }
    else
    {
        cerr << "Illegal SMT line: Target bracket not found." << endl;
        cerr << line << endl;
        return false;
    }

    return true;
}


bool MolReader::readSgroupDisplayInformation(Molecule *mol, string line)
{
    int num, scan;
    double x1, y1, x2, y2;

    scan = sscanf(line.c_str(), "M  SDI %d 4 %lf %lf %lf %lf",
                            &num, &x1, &y1, &x2, &y2);

    if(scan != 5)
    {
        cerr << "Illegal SDI line :" << endl;
        cerr << line << endl;
        return false;
    }

    Bracket *bracket = mol->getBracket(num -1);
    if(bracket)
    {
        Border *border;
        if(bracketDisplayInfo.count(bracket) == 0)
            border = bracket->getFirstBorder();
        else
            border = bracket->getSecondBorder();

        border->setFirstPoint(PointDbl(x1, y1));
        border->setSecondPoint(PointDbl(x2, y2));

        bracketDisplayInfo[bracket]++;
    }
    else
    {
        cerr << "Illegal SDI line :  Bracket not found." << endl;
        cerr << line << endl;
        return false;
    }

    return true;
}
