#include "Molecule.h"

using namespace std;
using namespace compound;

Molecule::Molecule()
{
}

Molecule::~Molecule()
{
    for(size_t i=0; i<atoms.size(); i++)
    {
        delete atoms[i];
    }

    for(size_t i=0; i<bonds.size(); i++)
    {
        delete bonds[i];
    }

    for(size_t i=0; i<brackets.size(); i++)
    {
        delete brackets[i];
    }
}


Atom * Molecule::addNewAtom()
{
    Atom *atom = new Atom();
    atoms.push_back(atom);
    return atom;
}


size_t Molecule::getAtomNumber() const
{
    return atoms.size();
}


Atom * Molecule::getAtom(size_t index) const
{
    Atom * ret = 0x00;

    if(index >= 0 && index < getAtomNumber())
    {
        ret = atoms[index];
    }

    return ret;
}


Bond *Molecule::addNewBond()
{
    Bond *bond = new Bond();
    bonds.push_back(bond);
    return bond;
}


size_t Molecule::getBondNumber() const
{
    return bonds.size();
}


Bond * Molecule::getBond(size_t index) const
{
    Bond *ret = 0x00;

    if(index >= 0 && index < getBondNumber())
    {
        ret = bonds[index];
    }

    return ret;
}


Bracket *Molecule::addNewBracket()
{
    Bracket *ret = new Bracket();
    brackets.push_back(ret);
    return ret;
}


size_t Molecule::getBracketNumber() const
{
    return brackets.size();
}


Bracket *Molecule::getBracket(size_t index) const
{
    Bracket *ret = 0x00;

    if(index >= 0 && index < getBracketNumber())
    {
        ret = brackets[index];
    }

    return ret;
}


void Molecule::move(double x, double y)
{
    moveAtom(x, y);
    moveBracket(x, y);
}


void Molecule::moveAtom(double x, double y)
{
    for(size_t i=0; i<getAtomNumber(); i++)
    {
        Atom *a = getAtom(i);
        a->setX(a->getX() + x);
        a->setY(a->getY() + y);
    }
}


void Molecule::moveBracket(double x, double y)
{
    utils::PointDbl p(x, y);

    for(size_t i=0; i<getBracketNumber(); i++)
    {
        getBracket(i)->move(x, y);
    }
}


void Molecule::scale(double times)
{
    scale(times, times);
}


void Molecule::scale(double xTimes, double yTimes)
{
    scaleAtom(xTimes, yTimes);
    scaleBracket(xTimes, yTimes);
}


void Molecule::scaleAtom(double xTimes, double yTimes)
{
    for(size_t i=0; i<getAtomNumber(); i++)
    {
        Atom *a = getAtom(i);
        a->setX(a->getX() * xTimes);
        a->setY(a->getY() * yTimes);
    }
}


void Molecule::scaleBracket(double xTimes, double yTimes)
{
    for(size_t i=0; i<getBracketNumber(); i++)
    {
        Bracket *bracket = getBracket(i);
        bracket->scale(xTimes, yTimes);
    }
}


void Molecule::setWidth(double width)
{
    double molWidth = getWidth();
    if(molWidth == 0.0) return;

    double times = width / molWidth;
    scale(times);
}


double Molecule::getWidth() const
{
    if(getAtomNumber() == 0)
        return 0;

    double minX, maxX;
    minX = DBL_MAX;
    maxX = -DBL_MAX;

    for(size_t i=0; i<getAtomNumber(); i++)
    {
        Atom *a = getAtom(i);

        if(a->getX() < minX)
            minX = a->getX();

        if(a->getX() > maxX)
            maxX = a->getX();

    }

    for(size_t i=0; i<getBracketNumber(); i++)
    {
        Bracket *bracket = getBracket(i);

        double bracketMinX = getBracketMinX(bracket);
        double bracketMaxX = getBracketMaxX(bracket);

        if(bracketMinX < minX) minX = bracketMinX;
        if(bracketMaxX > maxX) maxX = bracketMaxX;

    }

    return fabs(maxX - minX);
}


void Molecule::setHeight(double height)
{
    double molHeight = getHeight();
    if(molHeight == 0.0) return;

    double times = height / molHeight;
    scale(times);
}

double Molecule::getHeight() const
{
    if(getAtomNumber() == 0)
        return 0;

    double minY, maxY;
    minY = DBL_MAX;
    maxY = -DBL_MAX;

    for(size_t i=0; i<getAtomNumber(); i++)
    {
        Atom *a = getAtom(i);

        if(a->getY() < minY)
            minY = a->getY();

        if(a->getY() > maxY)
            maxY = a->getY();
    }

    for(size_t i=0; i<getBracketNumber(); i++)
    {
        Bracket *bracket = getBracket(i);

        double bracketMinY = getBracketMinY(bracket);
        double bracketMaxY = getBracketMaxY(bracket);

        if(bracketMinY < minY) minY = bracketMinY;
        if(bracketMaxY > maxY) maxY = bracketMaxY;
    }

    return fabs(maxY - minY);
}


double Molecule::getAverageBondLength() const
{
    if(getBondNumber() == 0)
        return getAverageAtomSpace();

    double sumLength = 0.0;
    for(size_t i=0; i<getBondNumber(); i++)
    {
        Bond *bond = getBond(i);
        sumLength += bond->getLength();
    }

    return (sumLength / getBondNumber());
}


double Molecule::getAverageAtomSpace() const
{
    if(getAtomNumber() <= 1)
        return 0.0;

    double sumSpace = 0;
    for(size_t i=0; (i+1)<getAtomNumber(); i++)
    {
        Atom *atom1 = getAtom(i);

        for(size_t j=(i+1); j<getAtomNumber(); j++)
        {
            Atom *atom2 = getAtom(j);

            double x = atom1->getX() - atom2->getX();
            double y = atom1->getY() - atom2->getY();

            sumSpace += sqrt(pow(x, 2) + pow(y, 2));
        }
    }

    return (sumSpace / (getAtomNumber()-1));
}


void Molecule::setBondLength(double length)
{
    double average = getAverageBondLength();
    if(average == 0.0)
        return;

    double times = length / average;
    scale(times);
}


void Molecule::setDimension(utils::DimensionDbl d)
{
    double molWidth = getWidth();
    double molHeight = getHeight();

    double times;
    if(molWidth > 0 && molHeight > 0)
    {
        double xTimes = d.getWidth() / molWidth;
        double yTimes = d.getHeight() / molHeight;

        times = (xTimes < yTimes) ? xTimes : yTimes;
    }
    else if(molWidth > 0)
    {
        times = d.getWidth() / molWidth;
    }
    else if(molHeight > 0)
    {
        times = d.getHeight() / molHeight;
    }
    else
    {
        return;
    }

    scale(times);
}


vector<Ring> Molecule::getRing()
{
    SimpleRingChecker ringChecker;
    vector<Ring> rings = ringChecker.getRing(this);

    if(rings.size() == 0)
        return rings;

    for(size_t i=0; i<(rings.size()-1); i++)
    {
        for(size_t j=i+1; j<rings.size(); j++)
        {
            Ring ring1 = rings[i];
            Ring ring2 = rings[j];

            if(ring1.getBondNumber() < ring2.getBondNumber())
            {
                rings[i] = ring2;
                rings[j] = ring1;
            }
        }
    }

    return rings;
}


bool Molecule::isColored()
{
    for(size_t i=0; i<getAtomNumber(); i++)
    {
        if(getAtom(i)->getColor() != "")
            return true;
    }

    for(size_t i=0; i<getBondNumber(); i++)
    {
        if(getBond(i)->getColor() != "")
            return true;
    }

    return false;
}
