#include "MoleculeImage.h"

using namespace std;
using namespace image;
using namespace compound;
using namespace utils;

const double MoleculeImage::AVERAGE_BOND_LENGTH = 30.0;
const double MoleculeImage::IMAGE_PADDING = 10.0;
const double MoleculeImage::DOUBLE_BOND_SPACE = 2.0;
const double MoleculeImage::TRIPLE_BOND_SPACE = 4.0;
const double MoleculeImage::STEREO_BOND_BASE = 4.0;
const double MoleculeImage::HASHED_BOND_SPACE = 3.0;

MoleculeImage::MoleculeImage(Image *image)
{
    this->image = image;
    isPractice = false;
}

void MoleculeImage::drawMolecule(Molecule *mol)
{
    RectangleDbl rect(0, 0);
    rect.setWidth(image->getWidth());
    rect.setHeight(image->getHeight());

    drawMolecule(mol, rect);
}


void MoleculeImage::drawMolecule(Molecule *mol, RectangleDbl rect)
{
    drawnBond.clear();

    RectangleDbl drawingRect = resizeDrawingRectangle(mol, rect);
    scaleMolecule(mol, drawingRect);
    moveOrigin(mol, drawingRect);

    drawRings(mol);
    drawBonds(mol);
    drawAtoms(mol);
    drawBrackets(mol);
}


bool MoleculeImage::isShowHydrogen()
{
    AtomEditMode mode = getAtomEditMode();

    bool ret = false;

    if(mode == Normal) ret = true;

    return ret;
}


string MoleculeImage::getAtomSymbol(Atom *atom)
{
    AtomEditMode mode = getAtomEditMode();

    string ret;
    if(mode == KCF)
    {
        ret = atom->getType();
    }
    else if(mode == SymbolAndNumber)
    {
        ostringstream oss;
        oss << atom->getSymbol();
        oss << atom->getNumber();

        ret = oss.str();
    }
    else
    {
        ret = atom->getSymbol();
    }

    return ret;
}


RectangleDbl MoleculeImage::getMoleculeRegion(const Molecule *mol)
{
    if(mol->getAtomNumber() == 0)
        return RectangleDbl(0, 0, 0, 0);

    isPractice = true;
    RectangleDbl ret = drawAtom(mol->getAtom(0));
    for(size_t i=1; i<mol->getAtomNumber(); i++)
    {
        //ret += getAtomSymbolRegion(mol->getAtom(i));
        ret += drawAtom(mol->getAtom(i));
    }
    isPractice = false;

    BracketImage brImage(image);
    for(size_t i=0; i<mol->getBracketNumber(); i++)
    {
        ret += brImage.getBracketRegion(mol->getBracket(i));

    }

    return ret;
}


bool MoleculeImage::isDisplayAtomSymbol(Atom *atom)
{
    AtomEditMode mode = getAtomEditMode();

    if(atom->isRepeatAtom())
    {
        return false;
    }
    else if(mode == KCF || mode == SymbolAndNumber || mode == SymbolOnly)
    {
        return true;
    }
    else if(mode == Normal)
    {
        if(atom->getBondNumber() <= 1)
            return true;
    }

    string symbol = getAtomSymbol(atom);

    if(symbol == "C")
    {
        if(atom->getColor() != "")
        {
            for(size_t i=0; i<atom->getBondNumber(); i++)
            {
                if(atom->getBond(i)->getColor() != "")
                {
                    return false;
                }
            }
        }
        else
        {
            if(atom->getBondNumber() > 1)
            {
                return false;
            }
        }
    }

    return true;
}


RectangleDbl MoleculeImage::resizeDrawingRectangle(Molecule *mol, RectangleDbl rect)
{
    double x1 = rect.getLeftTop().getX() + IMAGE_PADDING;
    double y1 = rect.getLeftTop().getY() + IMAGE_PADDING;
    double x2 = rect.getRightBottom().getX() - IMAGE_PADDING;
    double y2 = rect.getRightBottom().getY() - IMAGE_PADDING;

    return RectangleDbl(x1, y1, (x2-x1), (y2-y1));
}


void MoleculeImage::scaleMolecule(Molecule *mol, RectangleDbl rect)
{
    if(mol->getAtomNumber() <= 1)
        return;

    RectangleDbl molRect = getMoleculeRegion(mol);

    double atomSymbolWidth = molRect.getWidth() - mol->getWidth();
    double atomSymbolHeight = molRect.getHeight() - mol->getHeight();

    double drawingWidth = rect.getWidth() - atomSymbolWidth;
    double drawingHeight = rect.getHeight() - atomSymbolHeight;

    double molWidth = mol->getWidth();
    double molHeight = mol->getHeight();

    double scale1 = (drawingWidth > 0 && molWidth != 0) ? drawingWidth / molWidth : 1;
    double scale2 = (drawingHeight > 0 && molHeight != 0) ? drawingHeight / molHeight : 1;

    double scale = (scale1 < scale2) ? scale1 : scale2;

    mol->scale(scale);

    scaleBondSpace(mol);
}


void MoleculeImage::scaleBondSpace(Molecule *mol)
{
    double bondLength = mol->getAverageBondLength();

    if(bondLength > AVERAGE_BOND_LENGTH)
    {
        doubleBondSpace = DOUBLE_BOND_SPACE;
        tripleBondSpace = TRIPLE_BOND_SPACE;
        stereoBondBase = STEREO_BOND_BASE;
        return;
    }

    doubleBondSpace = DOUBLE_BOND_SPACE * bondLength / AVERAGE_BOND_LENGTH;
    if(doubleBondSpace < 1.0) doubleBondSpace = 1.0;

    tripleBondSpace = TRIPLE_BOND_SPACE * bondLength / AVERAGE_BOND_LENGTH;
    if(tripleBondSpace < 2.0) tripleBondSpace = 2.0;

    stereoBondBase = STEREO_BOND_BASE * bondLength / AVERAGE_BOND_LENGTH;
    if(stereoBondBase < 2.0) stereoBondBase = 2.0;
}


void MoleculeImage::moveOrigin(Molecule *mol, RectangleDbl rect)
{
    RectangleDbl molRect = getMoleculeRegion(mol);

    PointDbl leftTop = rect.getLeftTop();
    PointDbl molLeftTop = molRect.getLeftTop();

    double halfSpaceWidth = (rect.getWidth() - molRect.getWidth()) / 2;
    double halfSpaceHeight = (rect.getHeight() - molRect.getHeight()) / 2;

    double x = -molLeftTop.getX() + leftTop.getX() + halfSpaceWidth;
    double y = -molLeftTop.getY() + leftTop.getY() + halfSpaceHeight;
    mol->move(x, y);
}


void MoleculeImage::drawAtoms(Molecule *mol)
{
    for(size_t i=0; i<mol->getAtomNumber(); i++)
    {
        Atom *a = mol->getAtom(i);

        drawAtom(a);
    }
}


RectangleDbl MoleculeImage::drawAtom(Atom *atom)
{
    string s = getAtomSymbol(atom);

    if(!isDisplayAtomSymbol(atom))
        return RectangleDbl(atom->getX(), atom->getY());

    double x = atom->getX();
    double y = atom->getY();
    RGB rgb(atom->getColor());

    RectangleDbl rect;

    if(getAtomEditMode() == KCF || getAtomEditMode() == SymbolAndNumber)
    {
        rect = drawString(x, y, s, rgb, ALIGN_CENTER, VALIGN_CENTER, SMALL_FONT);
    }
    else
    {
        rect = drawString(x, y, s, rgb, ALIGN_CENTER, VALIGN_CENTER, LARGE_FONT);
    }

    rect = drawIsotope(atom, rect);

    rect = drawHydrogen(atom, rect);

    rect = drawCharge(atom, rect);

    return rect;
}


RectangleDbl MoleculeImage::drawIsotope(Atom *atom, RectangleDbl symbolRect)
{
    if(getAtomEditMode() == SymbolOnly)
        return symbolRect;

    int isotope = atom->getIsotope();

    if(isotope == 0) return symbolRect;

    // isotope string.
    ostringstream oss;
    oss << isotope;

    // foreground color.
    RGB rgb(atom->getColor());

    //
    PointDbl leftTop = symbolRect.getLeftTop();

    RectangleDbl rect = drawString(leftTop.getX(), leftTop.getY(), oss.str(),
                            rgb, ALIGN_RIGHT, VALIGN_CENTER, SMALL_FONT);

    return (symbolRect + rect);
}


RectangleDbl MoleculeImage::drawHydrogen(Atom *atom, RectangleDbl symbolRect)
{
    if(!isShowHydrogen()) return symbolRect;

    int hydrogenNumber = atom->getHydrogenNumber();
    if(hydrogenNumber == 0) return symbolRect;

    // atom color
    RGB rgb = RGB(atom->getColor());

    ostringstream oss;
    oss << hydrogenNumber;
    string coeff = oss.str();

    int hydrogenPosition = getHydrogenPosition(atom);

    RectangleDbl rect;
    if(hydrogenPosition == 1)
    {
        PointDbl rightBottom = symbolRect.getRightBottom();
        rect = drawString(
                rightBottom.getX(), rightBottom.getY(), "H", rgb,
                ALIGN_LEFT, VALIGN_BOTTOM, LARGE_FONT);

        if(hydrogenNumber > 1)
        {
            rightBottom = rect.getRightBottom();
            rect = drawString(
                    rightBottom.getX(), rightBottom.getY(), coeff, rgb,
                    ALIGN_LEFT, VALIGN_CENTER, SMALL_FONT);
        }
    }
    else if(hydrogenPosition == -1)
    {

        if(hydrogenNumber > 1)
        {
            PointDbl leftBottom = symbolRect.getLeftBottom();
            rect = drawString(
                    leftBottom.getX(), leftBottom.getY(), coeff, rgb,
                    ALIGN_RIGHT, VALIGN_CENTER, SMALL_FONT);
        }

        PointDbl leftBottom = symbolRect.getLeftBottom();
        rect = drawString(
                    leftBottom.getX()-rect.getWidth(), leftBottom.getY(), "H", rgb,
                    ALIGN_RIGHT, VALIGN_BOTTOM, LARGE_FONT);
    }

    return (symbolRect + rect);
}


RectangleDbl MoleculeImage::drawCharge(Atom *atom, RectangleDbl symbolRect)
{
    if(getAtomEditMode() == KCF || getAtomEditMode() == SymbolOnly)
        return symbolRect;

    int charge = atom->getCharge();

    if(charge == 0)
        return symbolRect;

    ostringstream oss;
    if(abs(charge) > 1)
        oss << abs(charge);

    if(charge > 0)
        oss << "+";
    else
        oss << "-";

    RGB rgb(atom->getColor());

    PointDbl rightTop = symbolRect.getRightTop();

    RectangleDbl rect = drawString(
                            rightTop.getX(), rightTop.getY(), oss.str(), rgb,
                            ALIGN_LEFT, VALIGN_CENTER, SMALL_FONT);

    return (symbolRect + rect);
}


RectangleDbl MoleculeImage::drawString(
        double x, double y, string str, RGB rgb,
        Align align, VAlign valign, FontType fontType)
{
    RectangleDbl ret;
    if(!isPractice)
    {
        RectangleDbl rect = image->getStringRegion(x, y, str, align, valign, fontType);
        image->filledRectangle(rect, RGB("#ffffff"));

        ret = image->drawString(x, y, str, rgb, align, valign, fontType);
    }
    else
    {
        ret = image->getStringRegion(x, y, str, align, valign, fontType);
    }

    return ret;
}


/*
 * return
 *    1 : hydrogen is right.
 *   -1 : hydrogen is left.
 *    0 : no hydrogen.
 *
 */
int MoleculeImage::getHydrogenPosition(Atom *atom)
{
    int ret = 1;

    if(atom->getBondNumber() == 1)
    {
        Bond *bond = atom->getBond(0);
        Atom *partner = bond->getConnectedAtom(atom);
        ret = (round(atom->getX()) >= round(partner->getX())) ? 1 : -1;
    }
    else if(atom->getBondNumber() == 0)
    {
        if(atom->getSymbol() == "O")
            return -1;
    }

    return ret;
}


void MoleculeImage::drawRings(Molecule *mol)
{
    vector<Ring> rings = mol->getRing();

    for(size_t i=0; i<rings.size(); i++)
    {
        drawRing(rings[i]);
    }
}


void MoleculeImage::drawRing(Ring ring)
{
    for(size_t i=0; i<ring.getBondNumber(); i++)
    {
        Bond *bond = ring.getBond(i);

        if(bond->isRepeatBond())
            continue;

        if(bond->isDouble())
        {
            drawRingDoubleBond(ring, bond);
        }
        else
        {
            drawBond(bond);
        }

        drawnBond[bond] = true;
    }
}


void MoleculeImage::drawRingDoubleBond(Ring ring, Bond *bond)
{
    if(drawnBond.count(bond) > 0)
        return;

    drawSingleBond(bond);
    drawInnerLine(ring, bond);
}


void MoleculeImage::drawInnerLine(Ring ring, Bond *bond)
{
    Atom *first = bond->getFirstAtom();
    Atom *second = bond->getSecondAtom();

    PointDbl center = ring.getCenter();

    PointDbl p1 = getInnerLineEndPoint(center, first);
    PointDbl p2 = getInnerLineEndPoint(center, second);

    RGB rgb(bond->getColor());

    image->drawLine(p1, p2, rgb);
}


PointDbl MoleculeImage::getInnerLineEndPoint(PointDbl center, Atom *atom)
{
    double space = doubleBondSpace * 3.0f;

    PointDbl ret(atom->getX(), atom->getY());

    if(fabs(center.getX() - atom->getX()) <= 1.0)
    {
        double y = (atom->getY() < center.getY()) ?
                        atom->getY() + space :
                        atom->getY() - space ;

        ret = PointDbl(atom->getX(), y);
    }
    else if(fabs(center.getY() - atom->getY()) <= 1.0)
    {
        double x = (atom->getX() < center.getX()) ?
                        atom->getX() + space :
                        atom->getX() - space ;

        ret = PointDbl(x, atom->getY());
    }
    else
    {
        double a = fabs((center.getY() - atom->getY()) / (center.getX() - atom->getX()));

        double x = sqrt(pow(space,2) / (1 + pow(a,2)));
        double y = a * x;

        if(atom->getX() < center.getX())
            x += atom->getX();
        else
            x = atom->getX() - x;

        if(atom->getY() < center.getY())
            y += atom->getY();
        else
            y = atom->getY() - y;

        ret = PointDbl(x, y);
    }

    return ret;
}


void MoleculeImage::drawBonds(Molecule *mol)
{
    for(size_t i=0; i<mol->getBondNumber(); i++)
    {
        Bond *b = mol->getBond(i);
        drawBond(b);
    }
}


void MoleculeImage::drawBond(Bond *bond)
{
    if(bond->isRepeatBond())
        return;

    if(drawnBond.count(bond) > 0)
        return;

    if(bond->isDouble())
        drawDoubleBond(bond);
    else if(bond->isTriple())
        drawTripleBond(bond);
    else if(bond->isUp())
        drawUpBond(bond);
    else if(bond->isDown())
        drawDownBond(bond);
    else
        drawSingleBond(bond);
}


void MoleculeImage::drawSingleBond(Bond *bond)
{
    Atom *first = bond->getFirstAtom();
    Atom *second = bond->getSecondAtom();

    double x1 = first->getX();
    double y1 = first->getY();
    double x2 = second->getX();
    double y2 = second->getY();
    RGB rgb(bond->getColor());

    image->drawLine(x1, y1, x2, y2, rgb);
}


void MoleculeImage::drawDoubleBond(Bond *bond)
{
    drawTwoLine(bond, doubleBondSpace);
}


void MoleculeImage::drawTripleBond(Bond *bond)
{
    drawSingleBond(bond);
    drawTwoLine(bond, tripleBondSpace);
}


void MoleculeImage::drawUpBond(Bond *bond)
{
    RGB rgb = RGB(bond->getColor());
    vector<PointDbl > points = calcStereoBondPoints(bond);
    image->filledPolygon(points, rgb);
}


void MoleculeImage::drawDownBond(Bond *bond)
{
    RGB rgb = RGB(bond->getColor());
    vector<PointDbl > points = calcStereoBondPoints(bond);

    int hashNumber = (int)round(bond->getLength() / (double)HASHED_BOND_SPACE);
    if(hashNumber <= 1)
        hashNumber = 2;

    vector<PointDbl > points1 = calcHashedBondPoints(points[0], points[1], hashNumber);
    vector<PointDbl > points2 = calcHashedBondPoints(points[0], points[2], hashNumber);

    for(size_t i=0; i<points1.size(); i++)
    {
        image->drawLine(points1[i], points2[i], rgb);
    }
}


void MoleculeImage::drawTwoLine(Bond *bond, double space)
{
    Atom *firstAtom = bond->getFirstAtom();
    Atom *secondAtom = bond->getSecondAtom();

    RGB clr(bond->getColor());

    PointDbl p = calcDoubleBondPoint(bond, space);

    double x1 = firstAtom->getX();
    double y1 = firstAtom->getY();
    double x2 = secondAtom->getX();
    double y2 = secondAtom->getY();

    if(x1 == x2)
    {
        // draw two vertical lines.
        image->drawLine(x1+p.getX(), y1, x2+p.getX(), y2, clr);
        image->drawLine(x1-p.getX(), y1, x2-p.getX(), y2, clr);
    }
    else if(y1 == y2)
    {
        // draw two horizontal lines.
        image->drawLine(x1, y1+p.getY(), x2, y1+p.getY(), clr);
        image->drawLine(x1, y1-p.getY(), x2, y1-p.getY(), clr);
    }
    else if((x1 < x2 && y1 < y2) || (x1 > x2 && y1 > y2))
    {
        /*
         *
         *   |                            |
         *   | *(x1, y1)                  | *(x2, y2)
         *   |                            |
         *   |       *(x2, y2)            |       *(x1, y1)
         *   |______________________      |______________________
         *
         */
        image->drawLine(x1+p.getX(), y1-p.getY(), x2+p.getX(), y2-p.getY(), clr);
        image->drawLine(x1-p.getX(), y1+p.getY(), x2-p.getX(), y2+p.getY(), clr);
    }
    else
    {
        /*
         *
         *   |                            |
         *   |       *(x1, y1)            |       *(x2, y2)
         *   |                            |
         *   | *(x2, y2)                  | *(x1, y1)
         *   |______________________      |______________________
         *
         */
        image->drawLine(x1+p.getX(), y1+p.getY(), x2+p.getX(), y2+p.getY(), clr);
        image->drawLine(x1-p.getX(), y1-p.getY(), x2-p.getX(), y2-p.getY(), clr);
    }
}


PointDbl MoleculeImage::calcDoubleBondPoint(Bond *bond, double space)
{
    Atom *first = bond->getFirstAtom();
    Atom *second = bond->getSecondAtom();

    double diffX = fabs((first->getX() - second->getX()));
    double diffY = fabs((first->getY() - second->getY()));

    double angle;
    if(diffX < (double)space)
        angle = M_PI_2;
    else
        angle = fabs(atan(diffY/diffX));

    angle = M_PI_2 - angle;

    double x = space * cos(angle);
    double y = space * sin(angle);

    PointDbl ret(x, y);

    return ret;
}


vector<PointDbl > MoleculeImage::calcStereoBondPoints(Bond *bond)
{
    Atom *first = bond->getFirstAtom();
    Atom *second = bond->getSecondAtom();

    PointDbl p = calcDoubleBondPoint(bond, stereoBondBase);

    double firstX = first->getX();
    double firstY = first->getY();
    double secondX = second->getX();
    double secondY = second->getY();

    double x2, y2, x3, y3;

    if(firstX == secondX)
    {
        x2 = secondX + p.getX();
        y2 = secondY;
        x3 = secondX - p.getX();
        y3 = secondY;
    }
    else if(firstY == secondY)
    {
        x2 = secondX;
        y2 = secondY - p.getY();
        x3 = secondX;
        y3 = secondY + p.getY();
    }
    else if((firstX < secondX && firstY < secondY) || (firstX > secondX && firstY > secondY))
    {
        /*
         *
         *   |                            |
         *   | *(firstX, firstY)          | *(secondX, secondY)
         *   |                            |
         *   |       *(secondX, secondY)  |       *(firstX, firstY)
         *   |______________________      |______________________
         *
         */
        x2 = secondX + p.getX();
        y2 = secondY - p.getY();
        x3 = secondX - p.getX();
        y3 = secondY + p.getY();
    }
    else
    {
        /*
         *
         *   |                            |
         *   |       *(firstX, firstY)    |       *(secondX, secondY)
         *   |                            |
         *   | *(secondX, secondY)        | *(firstX, firstY)
         *   |______________________      |______________________
         *
         */
        x2 = secondX + p.getX();
        y2 = secondY + p.getY();
        x3 = secondX - p.getX();
        y3 = secondY - p.getY();
    }

    vector<PointDbl > ret;
    ret.push_back(PointDbl(firstX, firstY));
    ret.push_back(PointDbl(x2, y2));
    ret.push_back(PointDbl(x3, y3));

    return ret;
}


vector<PointDbl > MoleculeImage::calcHashedBondPoints(PointDbl point1, PointDbl point2, int split)
{
    double x1, y1, x2, y2;
    x1 = point1.getX();
    y1 = point1.getY();
    x2 = point2.getX();
    y2 = point2.getY();

    vector<PointDbl > ret;
    if(x1 == x2)
    {
        double space = (y2 - y1) / (double)split;

        for(int i=0; i<split; i++)
        {
            double y = space * (double)(i+1) + y1;
            ret.push_back(PointDbl(x1, y));
        }
    }
    else
    {
        double coeff = (y2 - y1) / (x2 - x1);
        double space = (x2 - x1) / (double)split;

        for(int i=0; i<split; i++)
        {
            double x = space * (double)(i+1) + x1;
            double y = coeff * space * (double)(i+1) + y1;

            ret.push_back(PointDbl(x, y));
        }
    }

    return ret;
}


void MoleculeImage::drawBrackets(Molecule *mol)
{
    BracketImage bImage(image);

    for(size_t i=0; i<mol->getBracketNumber(); i++)
    {
        bImage.drawBracket(mol->getBracket(i));
    }
}


void MoleculeImage::setAtomEditMode(AtomEditMode mode)
{
    this->atomEditMode = mode;
}


AtomEditMode MoleculeImage::getAtomEditMode()
{
    return this->atomEditMode;
}
