#include "MoleculeKcfReader.h"

using namespace std;
using namespace compound;

bool MoleculeKcfReader::readKcf(Molecule *mol, string str)
{
    istringstream in(str);
    return read(mol, in);
}


bool MoleculeKcfReader::readKcfFile(Molecule *mol, string file)
{
    ifstream in;
    in.open(file.c_str(), ios::in);

    if(!in.good())
    {
        cerr << "Cound not open " << file << endl;
        return false;
    }

    return read(mol, in);
}

bool MoleculeKcfReader::read(Molecule *mol, istream &in)
{
    string line;

    string block = "";
    string bracket = "";

    char column[13];
    memset(column, 0x00, sizeof(column));

    while(getline(in, line))
    {
        if(line.find("ENTRY") == 0)
        {
        }
        else if(line.find("ATOM") == 0)
        {
            int atomNumber, scan;
            scan = sscanf(line.c_str(), "%12s%3d", column, &atomNumber);
            if(scan != 2)
            {
                cerr << "Illegal atom block:" << endl << line << endl;
                return false;
            }

            if(!readAtoms(mol, in, atomNumber))
            {
                return false;
            }
        }
        else if(line.find("BOND") == 0)
        {
            int bondNumber, scan;
            scan = sscanf(line.c_str(), "%12s%3d", column, &bondNumber);
            if(scan != 2)
            {
                cerr << "Illegal bond block:" << endl << line << endl;
                return false;
            }

            if(!readBonds(mol, in, bondNumber))
            {
                return false;
            }
        }
        else if(line.find("BRACKET") == 0)
        {
            block = "bracket";
        }
        else if(line.find(" ") != 0)
        {
            block = "";
        }
        else if(line.find("///") == 0)
        {
            break;
        }

        if(block == "bracket")
        {
            if(bracket != "") bracket += "\n";
            bracket += line;
        }
    }

    if(!readBracket(mol, bracket))
    {
        return false;
    }

    return true;
}


bool MoleculeKcfReader::readAtoms(Molecule *mol, istream &in, int atomNumber)
{
    string line;

    for(int i=0; i<atomNumber; i++)
    {
        if(!getline(in, line))
        {
            cerr << "Illegal atom block: atom not found. (index = " << (i+1) << ")" << endl;
            return false;
        }

        int num, scan;
        double x, y;
        char symbol[4], atomType[4];
        memset(symbol, 0x00, sizeof(symbol));
        memset(atomType, 0x00, sizeof(atomType));

        scan = sscanf(line.c_str(), "%3d %3s %2s %9lf %9lf",
                    &num, atomType, symbol, &x, &y);
        if(scan != 5)
        {
            cerr << "Illegal atom block:" << endl << line << endl;
            return false;
        }

        string color = getColor(line);

        Atom *atom = mol->addNewAtom();

        atom->setNumber((i+1));
        atom->setSymbol(string(symbol));
        atom->setType(string(atomType));
        atom->setX(x);
        atom->setY(y);
        atom->setCharge(getCharge(line));

        if(color != "")
            atom->setColor(color);
    }

    return true;
}


bool MoleculeKcfReader::readBonds(Molecule *mol, istream &in, int bondNumber)
{
    string line;

    for(int i=0; i<bondNumber; i++)
    {
        if(!getline(in, line))
        {
            cerr << "Illegal bond block: bond not found. (index = " << (i+1) << ")" << endl;
            return false;
        }

        int num, firstAtomNum, secondAtomNum, order, scan;
        bool up = false, down = false;

        scan = sscanf(line.c_str(), "%3d %3d %3d %d",
                    &num, &firstAtomNum, &secondAtomNum, &order);
        if(scan != 4)
        {
            cerr << "Illegal bond block:" << endl << line << endl;
            return false;
        }

        if(line.find("#Up") != string::npos)
        {
            up = true;
        }
        else if(line.find("#Down") != string::npos)
        {
            down = true;
        }

        string color = getColor(line);

        Atom *firstAtom, *secondAtom;
        firstAtom = mol->getAtom(firstAtomNum - 1);
        secondAtom = mol->getAtom(secondAtomNum - 1);

        if(!firstAtom || !secondAtom)
        {
            cerr << "Illegal bond block: atoms not found." << endl << line << endl;
            return false;
        }

        Bond *bond = mol->addNewBond();

        bond->setFirstAtom(firstAtom);
        bond->setSecondAtom(secondAtom);
        bond->setType(order);

        if(color != "")
            bond->setColor(color);

        firstAtom->addBond(bond);
        secondAtom->addBond(bond);

        if(up) bond->setUp();
        if(down) bond->setDown();
    }

    return true;
}


/*
 * example)
 *
 *  BRACKET     1    34.7200  -28.2800   34.7200  -23.4500
 *              1    40.2500  -23.4500   40.2500  -28.2800
 *              1  x
 *              2    23.2400  -30.0300   23.2400  -25.2700
 *              2    28.1400  -25.2700   28.1400  -30.0300
 *              2  3
 *    ORIGINAL  2  133 134 135 136 137
 *    REPEAT    2  138 139 140 141 142 143 144 145 146 147
 *
 */
bool MoleculeKcfReader::readBracket(Molecule *mol, string block)
{
    if(block.length() == 0) return true;

    istringstream is(block);

    string field = "";
    int bracketNum = 0;
    while(!is.eof())
    {
        int num, scan;
        string line1, line2, line3;

        getline(is, line1);

        if(line1.find("REPEAT") != string::npos)
        {
            field = "repeat";
        }

        line1 = line1.substr(12);

        // read bracket number.
        scan = sscanf(line1.c_str(), "%d", &num);

        if(bracketNum == num)
        {
            if(field == "repeat"){
                readRepeatAtom(mol, line1);
            }

            continue;
        }
        else
        {
            field = "";
            bracketNum = num;
        }

        getline(is, line2);
        getline(is, line3);

        line2 = line2.substr(12);
        line3 = line3.substr(12);

        double x1, y1, x2, y2, x3, y3, x4, y4;
        char subs[256];
        memset(subs, 0x00, sizeof(subs));

        // read first line.
        // "1     1.6138   -0.0483    1.6138    0.6828"
        scan = sscanf(line1.c_str(), "%d %lf %lf %lf %lf",
                        &num, &x1, &y1, &x2, &y2);
        if(scan != 5)
        {
            cerr << "Illegal bracket block:" << endl;
            cerr << line1 << endl;
            return false;
        }

        // read second line.
        // "1     4.1931    0.6828    4.1931   -0.0483"
        scan = sscanf(line2.c_str(), "%d %lf %lf %lf %lf",
                        &num, &x3, &y3, &x4, &y4);
        if(scan != 5)
        {
            cerr << "Illegal bracket block:" << endl;
            cerr << line2 << endl;
            return false;
        }

        // read third line.
        // "1  n"
        scan = sscanf(line3.c_str(), "%d %s", &num, subs);
        if(scan != 2)
        {
            cerr << "Illegal bracket block:" << endl;
            cerr << line3 << endl;
            return false;
        }

        // add bracket
        Bracket *bracket = mol->addNewBracket();

        Border *firstBorder = bracket->getFirstBorder();
        firstBorder->setFirstPoint(utils::PointDbl(x1, y1));
        firstBorder->setSecondPoint(utils::PointDbl(x2, y2));

        Border *secondBorder = bracket->getSecondBorder();
        secondBorder->setFirstPoint(utils::PointDbl(x3, y3));
        secondBorder->setSecondPoint(utils::PointDbl(x4, y4));

        bracket->setSubscript(string(subs));
    }

    return true;
}


bool MoleculeKcfReader::readRepeatAtom(Molecule *mol, string line)
{
    istringstream is(line);

    // first is bracket number;
    int bracketNum = 0;
    is >> bracketNum;

    // read repeat atom numbers.
    while(!is.eof())
    {
        int num = 0;
        is >> num;

        Atom *atom = mol->getAtom(num-1);
        if(atom != NULL)
        {
            atom->setRepeatAtom(true);
        }
    }

    return true;
}


string MoleculeKcfReader::getColor(const string str)
{
    istringstream is(str);

    string ret;
    while(!is.eof())
    {
        string s;
        is >> s;

        if(s.substr(0, 1) != "#") continue;

        // check string length
        if(s.length() != 7) continue;

        // check color format
        bool flag = true;
        for(size_t i=1; i<s.length(); i++)
        {
            char c = s.substr(i, 1).c_str()[0];
            if((c < '0' || c > '9') && (c < 'a' || c > 'f') && (c < 'A' || c > 'F'))
            {
                flag = false;
                break;
            }
        }

        if(!flag) continue;

        ret = s;
        break;
    }

    return ret;
}


int MoleculeKcfReader::getCharge(const string str)
{
    istringstream is(str);

    while(!is.eof())
    {
        string word;
        is >> word;

        if(word.find("#") != 0) continue;

        int charge;
        if(word == "#+")
            return 1;
        else if(word == "#-")
            return -1;
        else if(sscanf(word.c_str(), "#%d+", &charge) == 1)
            return charge;
        else if(sscanf(word.c_str(), "#%d-", &charge) == 1)
            return -charge;
    }

    return 0;
}
