#include "Glycan.h"

using namespace glycan;
using namespace compound;

Glycan::~Glycan()
{
    for(size_t i=0; i<nodes.size(); i++)
        delete nodes[i];

    for(size_t i=0; i<edges.size(); i++)
        delete edges[i];

    for(size_t i=0; i<brackets.size(); i++)
        delete brackets[i];
}


Node *Glycan::addNewNode()
{
    Node *node = new Node();
    if(node)
        nodes.push_back(node);
    else
        std::cerr << "Could not allocate a glycan::Node object." << std::endl;

    return node;
}


Node *Glycan::getNode(size_t index) const
{
    Node *ret = 0x00;

    if(index >= 0 && index < getNodeNumber())
    {
        ret = nodes[index];
    }

    return ret;
}


size_t Glycan::getNodeNumber() const
{
    return nodes.size();
}


Edge *Glycan::addNewEdge()
{
    Edge *edge = new Edge();

    if(edge)
        edges.push_back(edge);
    else
        std::cerr << "Could not allocate a glycan::Edge object." << std::endl;

    return edge;
}


Edge *Glycan::getEdge(size_t index) const
{
    Edge *ret = 0x00;

    if(index >= 0 && index < getEdgeNumber())
        ret = edges[index];

    return ret;
}


size_t Glycan::getEdgeNumber() const
{
    return edges.size();
}


Bracket *Glycan::addNewBracket()
{
    Bracket *ret = new Bracket();

    if(ret)
        brackets.push_back(ret);
    else
        std::cerr << "Could not allocate a Bracket." << std::endl;

    return ret;
}


Bracket *Glycan::getBracket(size_t index) const
{
    Bracket *ret;

    if(index >= 0 && index < getBracketNumber())
        ret = brackets[index];

    return ret;
}


size_t Glycan::getBracketNumber() const
{
    return brackets.size();
}


void Glycan::scale(double times)
{
    scale(times, times);
}


void Glycan::scale(double xTimes, double yTimes)
{
    scaleNode(xTimes, yTimes);
    scaleBracket(xTimes, yTimes);
}


void Glycan::scaleNode(double xTimes, double yTimes)
{
    for(size_t i=0; i<getNodeNumber(); i++)
    {
        Node *node = getNode(i);
        node->setX(node->getX() * xTimes);
        node->setY(node->getY() * yTimes);
    }
}


void Glycan::scaleBracket(double xTimes, double yTimes)
{
    for(size_t i=0; i<getBracketNumber(); i++)
    {
        getBracket(i)->scale(xTimes, yTimes);
    }
}


void Glycan::move(double x, double y)
{
    moveNode(x, y);
    moveBracket(x, y);
}


void Glycan::moveNode(double x, double y)
{
    for(size_t i=0; i<getNodeNumber(); i++)
    {
        Node *node = getNode(i);
        node->setX(node->getX() + x);
        node->setY(node->getY() + y);
    }
}


void Glycan::moveBracket(double x, double y)
{
    for(size_t i=0; i<getBracketNumber(); i++)
    {
        getBracket(i)->move(x, y);
    }
}


double Glycan::getAverageNodeSpace()
{
    if(getEdgeNumber() == 0)
        return 0.0;

    double sumSpace = 0;
    for(size_t i=0; i<getEdgeNumber(); i++)
    {
        Edge *edge = getEdge(i);

        Node *node1 = edge->getFirstNode();
        Node *node2 = edge->getSecondNode();

        sumSpace += fabs(node1->getX() - node2->getX());
    }

    return (sumSpace / getEdgeNumber());
}


void Glycan::setNodeSpace(double space)
{
    double averageNodeSpace = getAverageNodeSpace();
    if(averageNodeSpace == 0.0)
        return;

    double times = space / averageNodeSpace;

    scale(times);
}


void Glycan::setWidth(double width)
{
    double glWidth = getWidth();

    if(glWidth == 0) return;

    double times = width / glWidth;
    scale(times);
}


double Glycan::getWidth() const
{
    if(getNodeNumber() <= 1) return 0;

    double minX, maxX;
    minX = maxX = getNode(0)->getX();

    for(size_t i=1; i<getNodeNumber(); i++)
    {
        Node *node = getNode(i);

        if(node->getX() < minX)
            minX = node->getX();
        else if(node->getX() > maxX) 
            maxX = node->getX();
    }

    for(size_t i=0; i<getBracketNumber(); i++)
    {
        Bracket *bracket = getBracket(i);

        double bracketMinX = getBracketMinX(bracket);
        double bracketMaxX = getBracketMaxX(bracket);

        if(bracketMinX < minX)
            minX = bracketMinX;

        if(bracketMaxX > maxX)
            maxX = bracketMaxX;
    }

    return fabs(maxX - minX);
}


void Glycan::setHeight(double height)
{
    double glHeight = getHeight();

    if(glHeight == 0) return;

    double times = height / glHeight;
    scale(times);
}


double Glycan::getHeight() const
{
    if(getNodeNumber() <= 1) return 0;

    double minY, maxY;
    minY = maxY = getNode(0)->getY();

    for(size_t i=1; i<getNodeNumber(); i++)
    {
        Node *node = getNode(i);

        if(node->getY() < minY)
            minY = node->getY();
        else if(node->getY() > maxY)
            maxY = node->getY();
    }

    for(size_t i=0; i<getBracketNumber(); i++)
    {
        Bracket *bracket = getBracket(i);

        double bracketMinY = getBracketMinY(bracket);
        double bracketMaxY = getBracketMaxY(bracket);

        if(bracketMinY < minY)
            minY = bracketMinY;

        if(bracketMaxY > maxY)
            maxY = bracketMaxY;
    }

    return fabs(maxY - minY);
}


void Glycan::setDimension(utils::DimensionDbl d)
{
    double glWidth = getWidth();
    double glHeight = getHeight();

    double times;
    if(glWidth > 0 && glHeight > 0)
    {
        double xTimes = d.getWidth() / getWidth();
        double yTimes = d.getHeight() / getHeight();

        times = (xTimes < yTimes) ? xTimes : yTimes;
    }
    else if(glWidth > 0)
    {
        times = d.getWidth() / glWidth;
    }
    else if(glHeight > 0)
    {
        times = d.getHeight() / glHeight;
    }
    else
    {
        return;
    }

    scale(times);
}
