#include "GlycanImage.h"

using namespace std;
using namespace glycan;
using namespace image;
using namespace utils;
using namespace compound;

const double GlycanImage::IMAGE_PADDING = 10.0;
const double GlycanImage::NODE_MARGIN = 3.0;
const double GlycanImage::AVERAGE_NODE_SPACE = 90.0;

GlycanImage::GlycanImage(Image *image)
{
    this->image = image;
}


void GlycanImage::drawGlycan(Glycan *glycan)
{
    RectangleDbl rect(0, 0, image->getWidth(), image->getHeight());
    drawGlycan(glycan, rect);
}


void GlycanImage::drawGlycan(Glycan *glycan, RectangleDbl rect)
{
    RectangleDbl drawingRect = resizeDrawingRectangle(glycan, rect);

    scaleGlycan(glycan, drawingRect);
    moveOrigin(glycan, drawingRect);

    drawNodes(glycan);
    drawEdges(glycan);
    drawBracket(glycan);
}


RectangleDbl GlycanImage::getGlycanRegion(Glycan *glycan)
{
    if(glycan->getNodeNumber() == 0)
        return RectangleDbl(0, 0, 0, 0);

    // add node string size.
    RectangleDbl ret = getNodeRegion(glycan->getNode(0));
    for(size_t i=1; i<glycan->getNodeNumber(); i++)
    {
        ret += getNodeRegion(glycan->getNode(i));
    }

    // add bracket size.
    BracketImage brImage(image);
    for(size_t i=0; i<glycan->getBracketNumber(); i++)
    {
        ret += brImage.getBracketRegion(glycan->getBracket(i));
    }

    return ret;
}


RectangleDbl GlycanImage::getNodeRegion(const Node *node)
{
    return image->getStringRegion(
            node->getX(), node->getY(),
            node->getSymbol(),
            ALIGN_CENTER, VALIGN_CENTER);
}


RectangleDbl GlycanImage::resizeDrawingRectangle(Glycan *glycan, const RectangleDbl rect)
{
    double x1 = rect.getLeftTop().getX() + IMAGE_PADDING;
    double y1 = rect.getLeftTop().getY() + IMAGE_PADDING;
    double x2 = rect.getRightBottom().getX() - IMAGE_PADDING;
    double y2 = rect.getRightBottom().getY() - IMAGE_PADDING;

    return RectangleDbl(x1, y1, (x2-x1), (y2-y1));
}


void GlycanImage::scaleGlycan(Glycan *glycan, const RectangleDbl rect)
{
    if(glycan->getNodeNumber() <= 1)
        return;

    RectangleDbl glRect = getGlycanRegion(glycan);

    double drawingWidth = rect.getWidth();
    double drawingHeight = rect.getHeight();
    double glWidth = glRect.getWidth();
    double glHeight = glRect.getHeight();

    double scale1 = 1;
    if(drawingWidth > 0 && glWidth > 0){
        scale1 = drawingWidth / glWidth;
    }

    double scale2 = 1;
    if(drawingHeight > 0 && glHeight > 0){
        scale2 = drawingHeight / glHeight;
    }

    double scale = (scale1 < scale2) ? scale1 : scale2;
    glycan->scale(scale);
}


void GlycanImage::moveOrigin(Glycan *glycan, const RectangleDbl rect)
{
    RectangleDbl glRect = getGlycanRegion(glycan);

    PointDbl leftTop = rect.getLeftTop();
    PointDbl glLeftTop = glRect.getLeftTop();

    double halfSpaceWidth = (rect.getWidth() - glRect.getWidth()) / 2;
    double halfSpaceHeight = (rect.getHeight() - glRect.getHeight()) / 2;

    double x = -glLeftTop.getX() + leftTop.getX() + halfSpaceWidth;
    double y = -glLeftTop.getY() + leftTop.getY() + halfSpaceHeight;

    glycan->move((double)x, (double)y);
}


void GlycanImage::drawNodes(Glycan *glycan)
{
    for(size_t i=0; i<glycan->getNodeNumber(); i++)
    {
        Node *n = glycan->getNode(i);
        drawNode(n);
    }
}


void GlycanImage::drawNode(Node *node)
{

    string symbol = node->getSymbol();

    double x = node->getX();
    double y = node->getY();

    RGB rgb(node->getColor());

    image->drawString(x, y, symbol, rgb, ALIGN_CENTER, VALIGN_CENTER);
}


void GlycanImage::drawEdges(Glycan *glycan)
{
    for(size_t i=0; i<glycan->getEdgeNumber(); i++)
    {
        Edge *edge = glycan->getEdge(i);
        drawEdge(edge);
    }
}


void GlycanImage::drawEdge(Edge *edge)
{
    Node *firstNode = edge->getFirstNode();
    Node *secondNode = edge->getSecondNode();

    PointDbl firstPoint = calcEdgeStartPoint(firstNode, secondNode);
    PointDbl secondPoint = calcEdgeStartPoint(secondNode, firstNode);

    double x1 = firstPoint.getX();
    double y1 = firstPoint.getY();
    double x2 = secondPoint.getX();
    double y2 = secondPoint.getY();

    RGB rgb(edge->getColor());

    image->drawLine(x1, y1, x2, y2, rgb);

    drawHand(firstNode, edge);
    drawHand(secondNode, edge);
}


void GlycanImage::drawHand(Node *node, Edge *edge)
{
    string hand = node->getHand(edge);
    if(hand == "") return;

    Node *other = edge->getConnectedNode(node);

    PointDbl point = calcEdgeStartPoint(node, other);
    double x = point.getX();
    double y = point.getY();

    Align align;
    if(fabs(node->getX() - other->getX()) < 2.0)
    {
        align = ALIGN_CENTER;
    }
    else if(node->getX() < other->getX())
    {
        align = ALIGN_LEFT;
    }
    else
    {
        align = ALIGN_RIGHT;
    }

    VAlign valign;
    if(fabs(node->getY() - other->getY()) < 2.0)
    {
        valign = VALIGN_CENTER;
    }
    else if(node->getY() < other->getY())
    {
        valign = VALIGN_TOP;
    }
    else
    {
        valign = VALIGN_BOTTOM;
    }

    RGB rgb(edge->getColor());

    RectangleDbl rect = image->getStringRegion(x, y, hand, align, valign, SMALL_FONT);
    image->filledRectangle(rect, RGB(255,255,255));
    image->drawString(x, y, hand, rgb, align, valign, SMALL_FONT);
}


PointDbl GlycanImage::calcEdgeStartPoint(Node *firstNode, Node *secondNode)
{

    RectangleDbl firstRect = getNodeRegion(firstNode);
    RectangleDbl secondRect = getNodeRegion(secondNode);

    double firstNodeX = firstNode->getX();
    double firstNodeY = firstNode->getY();
    double secondNodeX = secondNode->getX();
    //double secondNodeY = secondNode->getY();

    double x, y;

    if(firstNodeX >= secondRect.getLeftTop().getX() &&
            firstNodeX <= secondRect.getRightTop().getX())
    {
        x = firstNodeX;
    }
    else if(firstNodeX < secondNodeX)
    {
        x = firstRect.getRightTop().getX() + NODE_MARGIN;
    }
    else
    {
        x = firstRect.getLeftTop().getX() - NODE_MARGIN;
    }

    if(firstNodeY >= secondRect.getLeftTop().getY() &&
            firstNodeY <= secondRect.getLeftBottom().getY())
    {
        y = firstNodeY;
    }
    else if(firstNodeY < secondRect.getLeftTop().getY())
    {
        y = firstRect.getRightBottom().getY() + NODE_MARGIN;
    }
    else
    {
        y = firstRect.getLeftTop().getY() - NODE_MARGIN;
    }

    return PointDbl(x, y);
}


void GlycanImage::drawBracket(Glycan *glycan)
{
    BracketImage brImage(image);

    for(size_t i=0; i<glycan->getBracketNumber(); i++)
    {
        brImage.drawBracket(glycan->getBracket(i));
    }
}
