#include "GlycanKcfReader.h"

using namespace std;
using namespace glycan;
using namespace compound;

bool GlycanKcfReader::readKcf(Glycan *gl, string str)
{
    istringstream in(str);
    return read(gl, in);
}


bool GlycanKcfReader::readKcfFile(Glycan *gl, string file)
{
    ifstream in;
    in.open(file.c_str(), ios::in);

    if(!in.good())
    {
        cerr << "Could not open " << file << endl;
        return false;
    }

    return read(gl, in);
}

bool GlycanKcfReader::read(Glycan *gl, istream &in)
{
    string line, block, bracket;

    char column[13];
    memset(column, 0x00, sizeof(column));

    while(getline(in, line))
    {
        if(line.find("ENTRY") == 0)
        {
        }
        else if(line.find("NODE") == 0)
        {
            int nodeNumber, scan;
            scan = sscanf(line.c_str(), "%12s%3d", column, &nodeNumber);
            if(scan != 2)
            {
                cerr << "Illegal node block:" << endl << line << endl;
                return false;
            }

            if(!readNodes(gl, in, nodeNumber))
            {
                return false;
            }
        }
        else if(line.find("EDGE") == 0)
        {
            int edgeNumber, scan;
            scan = sscanf(line.c_str(), "%12s%3d", column, &edgeNumber);
            if(scan != 2)
            {
                cerr << "Illegal edge block:" << endl << line << endl;
                return false;
            }

            if(!readEdges(gl, in, edgeNumber))
            {
                return false;
            }
        }
        else if(line.find("BRACKET") == 0)
        {
            block = "bracket";
        }
        else if(line.find("///") == 0)
        {
            break;
        }

        if(block == "bracket")
        {
            if(bracket != "") bracket += "\n";
            bracket += line;
        }
    }

    if(!readBracket(gl, bracket))
        return false;

    return true;
}


bool GlycanKcfReader::readNodes(Glycan *gl, istream &in, int nodeNumber)
{
    string line;

    for(int i=0; i<nodeNumber; i++)
    {
        if(!getline(in, line))
        {
            cerr << "Illegal node block: Node not found. (index = " << (i+1) << ")" << endl;
            return false;
        }

        int num, scan;
        double x, y;
        char symbol[128];
        memset(symbol, 0x00, sizeof(symbol));

        scan = sscanf(line.c_str(), "%3d %s %lf %lf",
                    &num, symbol, &x, &y);
        if(scan != 4)
        {
            cerr << "Illegal node block:" << endl << line << endl;
            return false;
        }

        string color = getColor(line);

        Node *node = gl->addNewNode();

        node->setSymbol(string(symbol));
        node->setX(x);
        node->setY(y);

        if(color != "")
            node->setColor(color);
    }

    return true;
}


bool GlycanKcfReader::readEdges(Glycan *gl, istream &in, int edgeNumber)
{
    string line;

    for(int i=0; i<edgeNumber; i++)
    {
        if(!getline(in, line))
        {
            cerr << "Illegal edge block: Edge not found. (index = " << (i+1) << ")" << endl;
            return false;
        }

        int num, firstNodeNum, secondNodeNum, scan;
        char firstNodeHand[4], secondNodeHand[4];
        char buff1[8], buff2[8];

        memset(buff1, 0x00, sizeof(buff1));
        memset(buff2, 0x00, sizeof(buff2));
        memset(firstNodeHand, 0x00, sizeof(firstNodeHand));
        memset(secondNodeHand, 0x00, sizeof(secondNodeHand));

        // read Edge line.
        // example)
        //   1     2:a1    1
        //   2     3:b1    2:4
        scan = sscanf(line.c_str(), "%3d %7s %7s",
                    &num, buff1, buff2);
        if(scan != 3)
            return false;

        // node number format is "1:a1" or "1"
        if(sscanf(buff1, "%d:%3s", &firstNodeNum, firstNodeHand) != 2)
        {
            if(sscanf(buff1, "%d", &firstNodeNum) != 1)
            {
                cerr << "Illegal edge block: First node number not found." << endl;
                cerr << line << endl;
                return false;
            }
        }

        if(sscanf(buff2, "%d:%3s", &secondNodeNum, secondNodeHand) != 2)
        {
            if(sscanf(buff2, "%d", &secondNodeNum) != 1)
            {
                cerr << "Illegal edge block: Second node number not found." << endl;
                cerr << line << endl;
                return false;
            }
        }

        string color = getColor(line);

        Node *firstNode, *secondNode;
        firstNode = gl->getNode(firstNodeNum - 1);
        secondNode = gl->getNode(secondNodeNum - 1);

        if(!firstNode || !secondNode)
        {
            cerr << "Illegal edge block: First node or second node not found." << endl;
            cerr << line << endl;
            return false;
        }

        Edge *edge = gl->addNewEdge();

        firstNode->addEdge(string(firstNodeHand), edge);
        secondNode->addEdge(string(secondNodeHand), edge);

        edge->setFirstNode(firstNode);
        edge->setSecondNode(secondNode);

        if(color != "")
            edge->setColor(color);
    }

    return true;
}


string GlycanKcfReader::getColor(const string str)
{
    istringstream is(str);

    string ret;
    while(!is.eof())
    {   
        string s;
        is >> s;

        if(s.substr(0, 1) != "#") continue;

        // check string length
        if(s.length() != 7) continue;

        // check color format
        bool flag = true;
        for(size_t i=1; i<s.length(); i++)
        {   
            char c = s.substr(i, 1).c_str()[0];
            if((c < '0' || c > '9') && (c < 'a' || c > 'f') && (c < 'A' || c > 'F'))
            {   
                flag = false;
                break;
            }   
        }   

        if(!flag) continue;

        ret = s;
        break;
    }   

    return ret;
}


bool GlycanKcfReader::readBracket(Glycan *gl, string block)
{
    if(block.length() == 0) return true;

    istringstream is(block);

    while(!is.eof())
    {   
        string line1, line2, line3;
        getline(is, line1);
        getline(is, line2);
        getline(is, line3);

        // remove column name. (BRACKET)
        if(line1.find("BRACKET") == 0)
            line1.replace(0, 7, "       ");

        int num, scan;
        double x1, y1, x2, y2, x3, y3, x4, y4; 
        char subs[256];
        memset(subs, 0x00, sizeof(subs));

        // read first line.
        // "1     1.6138   -0.0483    1.6138    0.6828"
        scan = sscanf(line1.c_str(), "%d %lf %lf %lf %lf",
                        &num, &x1, &y1, &x2, &y2);
        if(scan != 5)
        {   
            cerr << "Illegal bracket block:" << endl;
            cerr << line1 << endl;
            return false;
        }   

        // read second line.
        // "1     4.1931    0.6828    4.1931   -0.0483"
        scan = sscanf(line2.c_str(), "%d %lf %lf %lf %lf",
                        &num, &x3, &y3, &x4, &y4);
        if(scan != 5)
        {   
            cerr << "Illegal bracket block:" << endl;
            cerr << line2 << endl;
            return false;
        }   

        // read third line.
        // "1  n"
        scan = sscanf(line3.c_str(), "%d %s", &num, subs);
        if(scan != 2)
        {
            cerr << "Illegal bracket block:" << endl;
            cerr << line3 << endl;
            return false;
        }

        // add bracket
        Bracket *bracket = gl->addNewBracket();

        Border *firstBorder = bracket->getFirstBorder();
        firstBorder->setFirstPoint(utils::PointDbl(x1, y1));
        firstBorder->setSecondPoint(utils::PointDbl(x2, y2));

        Border *secondBorder = bracket->getSecondBorder();
        secondBorder->setFirstPoint(utils::PointDbl(x3, y3));
        secondBorder->setSecondPoint(utils::PointDbl(x4, y4));

        bracket->setSubscript(string(subs));
    }

    return true;
}

