#include <iostream>
#include <fstream>

#include "LigandFormat.h"
#include "../compound/MoleculeImage.h"
#include "../rpair/RpairImage.h"
#include "../glycan/GlycanImage.h"
#include "../image/ImageGD.h"
#include "../image/ImageFT.h"

#include "../utils/Utils.h"

#include "../compound/MoleculeKcfReader.h"
#include "../compound/MolReader.h"
#include "../rpair/RpairKcfReader.h"
#include "../glycan/GlycanKcfReader.h"

using namespace std;
using namespace utils;
using namespace image;
using namespace compound;
using namespace glycan;
using namespace rpair;

const int MOLECULE_FONT_SIZE = 14;
const int RPAIR_FONT_SIZE = 10;
const int GLYCAN_FONT_SIZE = 11;

extern char *optarg;
extern int optind, opterr, optopt;

typedef struct
{
    int width;
    int height;
    double connectionLength;
    string source;
    string sourceFile;
    string imageFile;
    string fontFile;
    int mode;
} Options;


void usage()
{
    cerr << "Usage : ligand_image -i source_file -o image_file" << endl;
    cerr << endl;
    cerr << "Option :" << endl;
    cerr << "  -i <file path>  KCF or MOL file path." << endl;
    cerr << "  -o <file path>  Output file path." << endl;
    cerr << endl;
    cerr << "  -f <file path>  TrueType Font path." << endl;
    cerr << endl;
    cerr << " Compound and Glycan" << endl;
    cerr << "  -w <width>      Image width." << endl;
    cerr << "  -h <height>     Image height." << endl;
    cerr << "  -l <length>     Average bond length." << endl;
    cerr << endl;
    cerr << " for experts" << endl;
    cerr << "  -m <mode>       Node print mode." << endl;
    cerr << "                    0 : Normal (default)" << endl;
    cerr << "                    1 : KCF" << endl;
    cerr << "                    2 : Symbol and Number" << endl;
    cerr << "                    3 : Symbol only" << endl;
    cerr << endl;
}


void getOptions(int argc, char *argv[], Options *o)
{
    char *optChars = "i:o:f:w:h:l:m:";

    o->width = 0;
    o->height = 0;
    o->connectionLength = 0.0;
    o->mode = 0;

    int option = getopt(argc, argv, optChars);

    while(option >= 0)
    {
        switch(option)
        {
            case 'i':
                o->sourceFile = string(optarg);
                break;
            case 'o':
                o->imageFile = string(optarg);
                break;
            case 'f':
                o->fontFile = string(optarg);
                break;
            case 'w':
                o->width = atoi(optarg);
                break;
            case 'h':
                o->height = atoi(optarg);
                break;
            case 'l':
                o->connectionLength = atof(optarg);
            case 'm':
                o->mode = atoi(optarg);
            default:
                break;
        }

        optarg = NULL;
        option = getopt(argc, argv, optChars);
    }
}


AtomEditMode getAtomEditMode(int i)
{
    AtomEditMode ret = Normal;

    switch (i)
    {
        case 1:
            ret = KCF;
            break;
        case 2:
            ret = SymbolAndNumber;
            break;
        case 3:
            ret = SymbolOnly;
            break;
        default:
            ret = Normal;
    }

    return ret;
}


image::Format getFormat(string file)
{
    if(lower(file).rfind(".png") == (file.length()-4))
        return image::PNG;
    else
        return image::GIF;
}


string getSource(istream &is)
{
    string ret, line;

    while(getline(is, line))
    {
        ret += line + "\n";
    }

    return ret;
}


DimensionInt getMoleculeDimension(Image *image, Molecule *mol)
{
    MoleculeImage molImage(image);
    RectangleDbl rect = molImage.getMoleculeRegion(mol);

    int width = (int)round(rect.getWidth() + MoleculeImage::IMAGE_PADDING * 2);
    int height = (int)round(rect.getHeight() + MoleculeImage::IMAGE_PADDING * 2);

    return DimensionInt(width, height);
}


void drawMoleculeImageGD(Molecule *mol, Options *options)
{
    ImageGD image;

    double padding = MoleculeImage::IMAGE_PADDING * 2;
    int width, height;
    if(options->width > 0 && options->height > 0)
    {
        mol->setDimension(DimensionDbl(options->width, options->height));

        width = options->width;
        height = options->height;
    }
    else if(options->width > 0)
    {
        mol->setWidth(options->width - padding);

        width = options->width;
        height = (int)(mol->getHeight() + padding);
    }
    else if(options->height > 0)
    {
        mol->setHeight(options->height - padding);

        width = (int)(mol->getWidth() + padding);
        height = options->height;
    }
    else
    {
        if(options->connectionLength > 0)
            mol->setBondLength(options->connectionLength);
        else
            mol->setBondLength(MoleculeImage::AVERAGE_BOND_LENGTH);

        DimensionInt d = getMoleculeDimension(&image, mol);
        width = d.getWidth();
        height  = d.getHeight();
    }

    image.init(width, height);

    MoleculeImage molImage(&image);
    molImage.setAtomEditMode(getAtomEditMode(options->mode));
    //if(options->mode == 1) molImage.setShowAtomType(true);
    //if(options->mode == 2) molImage.setShowAtomNumber(true);


    molImage.drawMolecule(mol);

    image.save(options->imageFile, getFormat(options->imageFile));
}


void drawMoleculeImageFT(Molecule *mol, Options *options)
{

    ImageFT image;
    image.setFontFile(options->fontFile);

    double padding = MoleculeImage::IMAGE_PADDING * 2;
    int width, height;
    if(options->width > 0 || options->height > 0 || options->connectionLength > 0)
    {
        if(options->connectionLength > 0)
        {
            mol->setBondLength(options->connectionLength);
        }
        else if(options->width > 0 && options->height > 0)
        {
            mol->setDimension(DimensionDbl(options->width, options->height));

            width = options->width;
            height = options->height;
        }
        else if(options->width > 0)
        {
            mol->setWidth(options->width - padding);

            width = options->width;
            height = (int)(mol->getHeight() + padding);
        }
        else
        {
            mol->setHeight(options->height - padding);

            width = (int)(mol->getWidth() + padding);
            height = options->height;
        }

#if 0
        int fontSize;
        if(options->connectionLength > 0 && mol->getAtomNumber() <= 1)
        {
            fontSize = MOLECULE_FONT_SIZE;
        }
        else
        {
            fontSize = (int)round(MOLECULE_FONT_SIZE *
                            mol->getAverageBondLength() / MoleculeImage::AVERAGE_BOND_LENGTH);
        }
#endif

        int fontSize = (int)round(MOLECULE_FONT_SIZE *
                            mol->getAverageBondLength() / MoleculeImage::AVERAGE_BOND_LENGTH);

        if(fontSize > MOLECULE_FONT_SIZE)
            fontSize = MOLECULE_FONT_SIZE;
        else if(fontSize < 6)
            fontSize = 6;

        image.setFontSize(fontSize);

        if(options->connectionLength > 0)
        {
            DimensionInt d = getMoleculeDimension(&image, mol);
            width = d.getWidth();
            height  = d.getHeight();
        }
    }
    else
    {
        mol->setBondLength(MoleculeImage::AVERAGE_BOND_LENGTH);
        image.setFontSize(MOLECULE_FONT_SIZE);

        DimensionInt d = getMoleculeDimension(&image, mol);
        width = d.getWidth();
        height  = d.getHeight();
    }

    image.init(width, height);

    MoleculeImage molImage(&image);
    molImage.setAtomEditMode(getAtomEditMode(options->mode));
    //if(options->mode == 1) molImage.setShowAtomType(true);
    //if(options->mode == 2) molImage.setShowAtomNumber(true);

    molImage.drawMolecule(mol);

    image.save(options->imageFile, getFormat(options->imageFile));
}


void drawMoleculeImage(Molecule *mol, Options *options)
{
    if(options->fontFile != "")
        drawMoleculeImageFT(mol, options);
    else
        drawMoleculeImageGD(mol, options);
}


void drawRpairImage(Rpair *rpair, Options *options)
{
    Image *image;

    if(options->fontFile != "")
    {
        ImageFT *i = new ImageFT();
        i->setFontFile(options->fontFile);
        i->setFontSize(RPAIR_FONT_SIZE);

        image = i;
    }
    else
    {
        image = new ImageGD();
    }

    Molecule *mol1 = rpair->getFirstMolecule();
    Molecule *mol2 = rpair->getSecondMolecule();

    mol1->setBondLength(RpairImage::AVERAGE_BOND_LENGTH);
    mol2->setBondLength(RpairImage::AVERAGE_BOND_LENGTH);

    AtomEditMode mode = getAtomEditMode(options->mode);
    MoleculeImage molImage(image);
    molImage.setAtomEditMode(mode);

    RectangleDbl rect1 = molImage.getMoleculeRegion(mol1);
    RectangleDbl rect2 = molImage.getMoleculeRegion(mol2);

    double width = rect1.getWidth() + rect2.getWidth() +
                    RpairImage::MOLECULE_SPACE +
                    MoleculeImage::IMAGE_PADDING * 4;

    double height = (rect1.getHeight() > rect2.getHeight()) ?
                    rect1.getHeight() : rect2.getHeight();
    height += MoleculeImage::IMAGE_PADDING * 2;

    image->init((int)width, (int)height);

    RpairImage rpairImage(image);
    rpairImage.setAtomEditMode(mode);
    rpairImage.drawRpair(rpair);

    image->save(options->imageFile, getFormat(options->imageFile));

    if(image) delete image;
}


DimensionInt getGlycanDimension(Image *image, Glycan *glycan)
{
    GlycanImage glImage(image);

    RectangleDbl rect = glImage.getGlycanRegion(glycan);

    int width = (int)round(rect.getWidth() + GlycanImage::IMAGE_PADDING * 2);
    int height = (int)round(rect.getHeight() + GlycanImage::IMAGE_PADDING * 2);

    return DimensionInt(width, height);
}


void drawGlycanImageGD(Glycan *glycan, Options *options)
{
    ImageGD image;

    double padding = GlycanImage::IMAGE_PADDING * 2;
    int width, height;
    if(options->width > 0 && options->height > 0)
    {
        glycan->setDimension(DimensionDbl(options->width, options->height));

        width = options->width;
        height = options->height;
    }
    else if(options->width > 0)
    {
        glycan->setWidth(options->width - padding);

        width = options->width;
        height = (int)(glycan->getHeight() + padding);
    }
    else if(options->height > 0)
    {
        glycan->setHeight(options->height - padding);

        width = (int)(glycan->getWidth() + padding);
        height = options->height;
    }
    else
    {
        glycan->setNodeSpace(GlycanImage::AVERAGE_NODE_SPACE);

        DimensionInt d = getGlycanDimension(&image, glycan);
        width = d.getWidth();
        height = d.getHeight();
    }

    image.init(width, height);

    GlycanImage glImage(&image);

    glImage.drawGlycan(glycan);

    image.save(options->imageFile, getFormat(options->imageFile));
}


void drawGlycanImageFT(Glycan *glycan, Options *options)
{
    ImageFT image;
    image.setFontFile(options->fontFile);

    double padding = GlycanImage::IMAGE_PADDING * 2;
    int width, height;
    if(options->width > 0 || options->height > 0)
    {
        if(options->width > 0 && options->height > 0)
        {
            glycan->setDimension(DimensionDbl(options->width, options->height));

            width = options->width;
            height = options->height;
        }
        else if(options->width > 0)
        {
            glycan->setWidth(options->width - padding);

            width = options->width;
            height = (int)(glycan->getHeight() + padding);
        }
        else
        {
            glycan->setHeight(options->height - padding);

            width = (int)(glycan->getWidth() + padding);
            height = options->height;
        }

        int fontSize = (int)round(GLYCAN_FONT_SIZE *
                            glycan->getAverageNodeSpace() / GlycanImage::AVERAGE_NODE_SPACE);

        if(fontSize > GLYCAN_FONT_SIZE)
            fontSize = GLYCAN_FONT_SIZE;
        else if(fontSize < 6)
            fontSize = 6;

        image.setFontSize(fontSize);
    }
    else
    {
        image.setFontSize(GLYCAN_FONT_SIZE);

        glycan->setNodeSpace(GlycanImage::AVERAGE_NODE_SPACE);

        DimensionInt d = getGlycanDimension(&image, glycan);
        width = d.getWidth();
        height = d.getHeight();
    }

    image.init(width, height);

    GlycanImage glImage(&image);
    glImage.drawGlycan(glycan);

    image.save(options->imageFile, getFormat(options->imageFile));
}

void drawGlycanImage(Glycan *glycan, Options *options)
{
    if(options->fontFile != "")
        drawGlycanImageFT(glycan, options);
    else
        drawGlycanImageGD(glycan, options);
}


int main(int argc, char *argv[])
{

    Options options;
    getOptions(argc, argv, &options);

    if(options.imageFile == "")
    {
        usage();
        return -1;
    }


    // read structure data.
    if(options.sourceFile != "")
    {
        ifstream in;
        in.open(options.sourceFile.c_str(), ios::in);
        if(!in.good())
        {
            cerr << "Could not open " << options.sourceFile << "." << endl;
            return -1;
        }

        options.source = getSource(in);
        in.close();
    }
    else
    {
        options.source = getSource(cin);
    }


    // structure format check
    bool isMoleculeKCF = ligand::isMoleculeKCF(options.source);
    bool isRpairKCF = ligand::isRpairKCF(options.source);
    bool isGlycanKCF = ligand::isGlycanKCF(options.source);
    bool isMol = ligand::isMol(options.source);

    if(isMoleculeKCF)
    {
        Molecule mol;
        MoleculeKcfReader reader;

        if(!reader.readKcf(&mol, options.source))
            return -1;

        mol.scale(1, -1);

        drawMoleculeImage(&mol, &options);
    }
    else if(isMol)
    {
        Molecule mol;
        MolReader reader;

        if(!reader.readMol(&mol, options.source))
            return -1;

        mol.scale(1, -1);

        drawMoleculeImage(&mol, &options);
    }
    else if(isRpairKCF)
    {
        Rpair rpair;
        RpairKcfReader reader;

        if(!reader.readKcf(&rpair, options.source))
            return -1;

        rpair.getFirstMolecule()->scale(1, -1);
        rpair.getSecondMolecule()->scale(1, -1);

        drawRpairImage(&rpair, &options);
    }
    else if(isGlycanKCF)
    {
        Glycan glycan;
        GlycanKcfReader reader;

        if(!reader.readKcf(&glycan, options.source))
            return -1;

        glycan.scale(1, -1);

        for(size_t i=0; i<glycan.getEdgeNumber(); i++)
            glycan.getEdge(i)->setColor("#0000ff");

        drawGlycanImage(&glycan, &options);
    }
    else
    {
        cerr << "Unkonwn source format: ";
        if(options.sourceFile != "")
            cerr << options.sourceFile << endl;
        else
            cerr << "Standard input" << endl;

        return -1;
    }

    return 0;
}

