#include "RpairImage.h"

using namespace std;
using namespace image;
using namespace rpair;
using namespace compound;
using namespace utils;

const double RpairImage::MOLECULE_SPACE = 10.0;
const double RpairImage::AVERAGE_BOND_LENGTH = 35.0;

const string RpairImage::COLOR_REACTION_CENTER = "#ff0000";
const string RpairImage::COLOR_DIFFERENCE_ATOM = "#0000ff";
const string RpairImage::COLOR_MATCHED_ATOM = "#ff9900";
const string RpairImage::COLOR_PAIRED_ATOM = "#00ff00";


RpairImage::RpairImage(Image *image)
{
    this->image = image;
    setAtomEditMode(KCF);
}


void RpairImage::drawRpair(Rpair *rpair)
{
    setAlignmentColor(rpair);

    MoleculeImage molImage(this->image);
    molImage.setAtomEditMode(getAtomEditMode());

    RectangleDbl firstMolRect = getFirstMoleculeRegion(&molImage, rpair);
    RectangleDbl secondMolRect = getSecondMoleculeRegion(&molImage, rpair);

    molImage.drawMolecule(rpair->getFirstMolecule(), firstMolRect);
    molImage.drawMolecule(rpair->getSecondMolecule(), secondMolRect);
}


RectangleDbl RpairImage::getFirstMoleculeRegion(MoleculeImage *molImage, Rpair *rpair)
{
    Molecule *firstMol = rpair->getFirstMolecule();
    Molecule *secondMol = rpair->getSecondMolecule();

    RectangleDbl firstMolRect = molImage->getMoleculeRegion(firstMol);
    RectangleDbl secondMolRect = molImage->getMoleculeRegion(secondMol);

    double imageWidth = image->getWidth();
    double imageHeight = image->getHeight();

    double drawingWidth = imageWidth - MOLECULE_SPACE - MoleculeImage::IMAGE_PADDING * 4;

    double firstMolWidth = firstMolRect.getWidth();
    double secondMolWidth = secondMolRect.getWidth();

    double scale = firstMolWidth / (firstMolWidth + secondMolWidth);

    double molWidth = drawingWidth * scale + MoleculeImage::IMAGE_PADDING * 2;

    return RectangleDbl(0, 0, molWidth, imageHeight);
}


RectangleDbl RpairImage::getSecondMoleculeRegion(MoleculeImage *molImage, Rpair *rpair)
{
    Molecule *firstMol = rpair->getFirstMolecule();
    Molecule *secondMol = rpair->getSecondMolecule();

    RectangleDbl firstMolRect = molImage->getMoleculeRegion(firstMol);
    RectangleDbl secondMolRect = molImage->getMoleculeRegion(secondMol);

    double imageWidth = image->getWidth();
    double imageHeight = image->getHeight();

    double drawingWidth = imageWidth - MOLECULE_SPACE - MoleculeImage::IMAGE_PADDING * 4;

    double firstMolWidth = firstMolRect.getWidth();
    double secondMolWidth = secondMolRect.getWidth();

    double scale = secondMolWidth / (firstMolWidth + secondMolWidth);

    double molWidth = drawingWidth * scale + MoleculeImage::IMAGE_PADDING * 2;

    return RectangleDbl((imageWidth - molWidth), 0, molWidth, imageHeight);
}


void RpairImage::setAlignmentColor(Rpair *rpair)
{
    if(rpair->getFirstMolecule()->isColored())
        return;

    if(rpair->getSecondMolecule()->isColored())
        return;

    setAtomColor(rpair);

    setBondColor(rpair, rpair->getFirstMolecule());
    setBondColor(rpair, rpair->getSecondMolecule());
}


void RpairImage::setAtomColor(Rpair *rpair)
{
    for(int i=0; i<rpair->getAtomPairNumber(); i++)
    {
        AtomPair pair = rpair->getAtomPair(i);

        Atom *firstAtom = pair.getFirstAtom();
        Atom *secondAtom = pair.getSecondAtom();

        string color;
        if(pair.isReactionCenter())
        {
            color = COLOR_REACTION_CENTER;
        }
        else if(pair.isDifferenceAtom())
        {
            color = COLOR_DIFFERENCE_ATOM;
        }
        else if(pair.isMatchedAtom())
        {
            color = COLOR_MATCHED_ATOM;
        }
        else
        {
            color = COLOR_PAIRED_ATOM;
        }

        if(firstAtom && firstAtom->getColor() == "")
            firstAtom->setColor(color);

        if(secondAtom && secondAtom->getColor() == "")
            secondAtom->setColor(color);
    }
}


void RpairImage::setBondColor(Rpair *rpair, Molecule *mol)
{
    for(size_t i=0; i<mol->getBondNumber(); i++)
    {
        Bond *bond = mol->getBond(i);

        Atom *firstAtom = bond->getFirstAtom();
        Atom *secondAtom = bond->getSecondAtom();

        string color1 = firstAtom->getColor();
        string color2 = secondAtom->getColor();

        if(isDifferenceAtom(rpair, firstAtom) && isDifferenceAtom(rpair, secondAtom))
        {
            bond->setColor(COLOR_DIFFERENCE_ATOM);
        }
        else if(color1 == color2)
        {
            bond->setColor(color1);
        }
        else if(color1 == "#000000" || color2 == "#000000")
        {
            bond->setColor("#000000");
        }
        else if(color1 == COLOR_PAIRED_ATOM || color2 == COLOR_PAIRED_ATOM)
        {
            bond->setColor(COLOR_PAIRED_ATOM);
        }
        else if(color1 == COLOR_MATCHED_ATOM || color2 == COLOR_MATCHED_ATOM)
        {
            bond->setColor(COLOR_MATCHED_ATOM);
        }
        else if(color1 == COLOR_DIFFERENCE_ATOM || color2 == COLOR_DIFFERENCE_ATOM)
        {
            bond->setColor(COLOR_DIFFERENCE_ATOM);
        }
    }
}


bool RpairImage::isDifferenceAtom(Rpair *rpair, Atom *atom)
{
    for(int i=0; i<rpair->getAtomPairNumber(); i++)
    {
        AtomPair pair = rpair->getAtomPair(i);

        if(atom == pair.getFirstAtom() || atom == pair.getSecondAtom())
        {
            if(pair.isDifferenceAtom())
                return true;
        }
    }

    return false;
}


void RpairImage::setAtomEditMode(AtomEditMode mode)
{
    this->atomEditMode = mode;
}


AtomEditMode RpairImage::getAtomEditMode()
{
    return this->atomEditMode;
}
